@@ -6,6 +6,7 @@ import Prelude hiding (null, length)
 import Data.ByteString (ByteString)
 import qualified Data.ByteString as B
 import Criterion.Main
+import Data.Monoid
 import Data.Word
 import Data.Iteratee
 import Data.Iteratee.Base.ReadableChunk
@@ -42,9 +43,32 @@ testHdByte = fileDriverHandle bufSize len file >> return ()
   len :: Monad m => Iteratee ByteString m Int
   len = length
 
+testFdMapReduce :: Int -> IO ()
+testFdMapReduce n = fileDriverFd bufSize sum file >> return ()
+ where
+  sum :: Iteratee ByteString IO Word8
+  sum = getSum `fmap` mapReduce n (Sum . B.foldl' (+) 0)
+
+testFdFold :: IO ()
+testFdFold = fileDriverFd bufSize sum file >> return ()
+ where
+  sum :: Iteratee ByteString IO Word8
+  sum = foldl' (+) 0
+
 main = defaultMain
-  [ bench "Fd with String" testFdString
-  , bench "Hd with String" testHdString
-  , bench "Fd with ByteString" testFdByte
-  , bench "Hd with ByteString" testHdByte
+  [
+   bgroup "String" [
+     bench "Fd" testFdString
+    ,bench "Hd with String" testHdString
+   ]
+  ,bgroup "ByteString" [
+     bench "Fd" testFdByte
+    ,bench "Hd" testHdByte
+   ]
+  ,bgroup "folds" [
+     bench "Fd/fold" testFdFold
+    ,bench "Fd/mapReduce 2" $ testFdMapReduce 2
+    ,bench "Fd/mapReduce 4" $ testFdMapReduce 4
+    ,bench "Fd/mapReduce 8" $ testFdMapReduce 8
+   ]
   ]
